import logging
from exceptions import *
import xml.dom.minidom


class Feed(object):
    def __init__(self, doc):
        #construct dom doc if we were passed a string
        if isinstance(doc, basestring):
            logging.info("doc is a string. make it a dom")
            doc = xml.dom.minidom.parseString(doc)
            if not doc:
                logging.info(doc)
                raise ZoteroParseError("Error constructing feed Domdoc")
        feedEl = doc.getElementsByTagName('feed').item(0)
        if not feedEl:
            raise ZoteroParseError("No feed node in passed doc")
        self.title = feedEl.getElementsByTagName('title').item(0).childNodes.item(0).nodeValue
        self.id = feedEl.getElementsByTagName("id").item(0).childNodes.item(0).nodeValue
        self.dateUpdated = feedEl.getElementsByTagName("updated").item(0).childNodes.item(0).nodeValue
        self.totalResults = feedEl.getElementsByTagName("zapi:totalResults").item(0).childNodes.item(0).nodeValue
        #grab links of the feed element (and not all the entries)
        self.links = {}
        for n in feedEl.childNodes:
            if n.nodeName == 'link':
                linkrel = n.getAttribute('rel')
                linktype = n.getAttribute('type')
                linkhref = n.getAttribute('href')
                self.links[linkrel] = {'type': linktype, 'href': linkhref, 'rel': linkrel}

        self.entries = feedEl.getElementsByTagName('entry')



import time
import logging


class Cache(object):
    def __init__(self, backend='', ttl=300):
        self.store = {}
        self.expirations = {}
        self.ttl = ttl
        self.garbageInterval = 60  # how many operations between garbage collection
        self.opCount = 0  # count of operations for garbage collection purposes
        return

    def cache_store(self, key, value, lifetime=None):
        if lifetime == None:
            lifetime = self.ttl
        expiration = time.time() + lifetime
        self.store[key] = value
        self.expirations[key] = expiration
        logging.info("value stored in cache - " + key)
        #increment opCount and possibly garbage collect
        self.opCount += 1
        if self.opCount > self.garbageInterval:
            logging.info("sweeping cache")
            self._sweep_cache()
        return None

    def cache_fetch(self, key):
        now = time.time()
        if key in self.expirations:
            if self.expirations[key] < now:
                logging.info("Cache hit, but expired - returning None")
                return None
            else:
                if key in self.store:
                    logging.info('Cache hit - returning cached value')
                    return self.store[key]
        else:
            logging.info("Cache miss - returning None")
            return None

    def _sweep_cache(self):
        now = time.time()
        for key in self.expirations.keys():
            if self.expirations[key] < now:
                del self.store[key]
                del self.expirations[key]



#import logging
import json
from entry import *


class Collection(Entry):
    def __init__(self, entryNode=None, library=None):
        self.libraryUrlIdentifier = ''
        self.title = ''
        self.name = ''
        self.collectionKey = None
        self.collectionVersion = 0
        self.numCollections = 0
        self.numItems = 0
        self.parentCollectionKey = False
        self.apiObject = {}
        self.pristine = {}

        if library != None:
            self.associateWithLibrary(library)
        if entryNode == None:
            return
        super(Collection, self).__init__(entryNode)
        self.collectionKey = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', 'key').item(0).childNodes.item(0).nodeValue
        self.collectionVersion = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', 'version').item(0).childNodes.item(0).nodeValue
        self.numCollections = entryNode.getElementsByTagName('http://zotero.org/ns/api', 'numCollections').item(0).childNodes.item(0).nodeValue
        self.numItems = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', 'numItems').item(0).childNodes.item(0).nodeValue

        contentNode = entryNode.getElementsByTagName('content').item(0)
        if contentNode:
            contentType = contentNode.getAttribute('type')
            if contentType == 'application/json':
                jsonstring = contentNode.childNodes.item(0).nodeValue
                self.pristine = json.loads(jsonstring)
                self.apiObject = json.loads(jsonstring)
                self.parentCollectionKey = self.apiObject['parentCollection']
                self.name = self.apiObject['name']

    def writeApiObject(self):
        return dict(self.pristine.items() + self.apiObject.items())

    def get(self, key):
        if key == 'title' or key == 'name':
            return self.name
        elif key == 'collectionKey' or key == 'key':
            return self.collectionKey
        elif key == 'collectionVersion' or key == 'version':
            return self.collectionVersion
        elif key == 'parentCollection' or key == 'parentCollectionKey':
            return self.parentCollectionKey

        if key in self.apiObject:
            return self.apiObject[key]

        if key in dir(self):
            return self.__getattribute__(key)

        return None

    def set(self, key, val):
        if key in self.apiObject:
            self.apiObject[key] = val

        if key == 'title' or key == 'name':
            self.name = val
            self.apiObject['name'] = val
        elif key == 'collectionKey' or key == 'key':
            self.collectionKey = val
            self.apiObject['collectionKey'] = val
        elif key == 'parentCollection' or key == 'parentCollectionKey':
            self.parentCollectionKey = val
            self.apiObject['parentCollection'] = val
        elif key == 'collectionVersion' or key == 'version':
            self.collectionVersion = val
            self.apiObject['collectionVersion'] = val

        if key in self.apiObject:
            self.apiObject[key] = val

        if key in dir(self):
            self.__setattr__(key, val)



import sys
sys.path.append('../')
import unittest
from libZotero import zotero


class TestApiUrlBuilding(unittest.TestCase):

    def test_apiurls(self):
        config = {'target': 'collections', 'libraryType': 'user', 'libraryID': 1, 'content': 'json', 'limit': '100'}
        self.assertEqual(zotero.apiRequestUrl(config), "https://api.zotero.org/users/1/collections")
        self.assertEqual(zotero.apiQueryString(config), "?content=json&limit=100")

        config = {'target': 'items', 'libraryType': 'group', 'libraryID': 1, 'format': 'atom', 'content': 'json', 'limit': '100'}
        self.assertEqual(zotero.apiRequestUrl(config), "https://api.zotero.org/groups/1/items")
        self.assertEqual(zotero.apiQueryString(config), "?content=json&format=atom&limit=100")

        config = {'target': 'items', 'libraryType': 'user', 'libraryID': 1, 'content': 'json,coins', 'limit': '25'}
        self.assertEqual(zotero.apiRequestUrl(config), "https://api.zotero.org/users/1/items")
        self.assertEqual(zotero.apiQueryString(config), "?content=json%2Ccoins&limit=25")

        config = {'target': 'item', 'libraryType': 'user', 'libraryID': 1, 'content': 'json', 'itemKey': 'ASDF1234'}
        self.assertEqual(zotero.apiRequestUrl(config), "https://api.zotero.org/users/1/items/ASDF1234")
        self.assertEqual(zotero.apiQueryString(config), "?content=json")

        config = {'target': 'items', 'libraryType': 'user', 'libraryID': 1, 'content': 'bibtex', 'limit': '100', 'itemKey': 'ASDF1234,FDSA4321'}
        self.assertEqual(zotero.apiRequestUrl(config), "https://api.zotero.org/users/1/items")
        self.assertEqual(zotero.apiQueryString(config), "?content=bibtex&itemKey=ASDF1234%2CFDSA4321&limit=100")

        config = {'target': 'deleted', 'libraryType': 'user', 'libraryID': 1, 'content': 'json', 'limit': '100'}
        self.assertEqual(zotero.apiRequestUrl(config), "https://api.zotero.org/users/1/deleted")
        self.assertEqual(zotero.apiQueryString(config), "?content=json&limit=100")

        config = {'target': 'children', 'libraryType': 'user', 'libraryID': 1, 'itemKey': 'ASDF1234', 'content': 'json', 'limit': '100'}
        self.assertEqual(zotero.apiRequestUrl(config), "https://api.zotero.org/users/1/items/ASDF1234/children")
        self.assertEqual(zotero.apiQueryString(config), "?content=json&limit=100")


if __name__ == '__main__':
    unittest.main()



class ZoteroError(Exception):
    """Base class for exceptions in this module."""
    pass


class ZoteroUrlError(ZoteroError):
    pass


class ZoteroApiError(ZoteroError):
    pass


class ZoteroParseError(ZoteroError):
    pass



import logging
import xml.dom.minidom
import re
import json
from entry import *


class Tag(Entry):
    def __init__(self, entryNode):
        self.tagID = None
        self.libraryID = None
        self.key = None
        self.name = None
        self.dateAdded = None
        self.dateModified = None
        self.type = None
        self.numItems = 0

        if entryNode == None:
            return
        elif isinstance(entryNode, basestring):
            doc = xml.dom.minidom.parseString(entryNode)
            entryNode = doc.getElementsByTagName("entry").item(0)
        super(Tag, self).__init__(entryNode)
        self.name = self.title
        numItems = entryNode.getElementsByTagNameNS('*', "numItems").item(0).childNodes.item(0).nodeValue
        if(numItems):
            self.numItems = int(numItems)

        tagElements = entryNode.getElementsByTagName("tag")
        self.tagElement = tagElements.item(0)

    def dataObject(self):
        tagDict = {}
        tagDict.title = self.title
        tagDict.dateAdded = self.dateAdded
        tagDict.dateUpdated = self.dateUpdated
        tagDict.id = self.id
        tagDict.properties = self.properties

        return tagDict



import logging
from collection import *
from zotero import getKey, responseIsError, updateObjectsFromWriteResponse


class Collections(object):
    def __init__(self):
        self.orderedArray = []
        self.collectionObjects = {}
        self.dirty = False
        self.loaded = False
        self.owningLibrary = None
        self.collectionsVersion = 0

    def __len__(self):
        return len(self.collectionObjects)

    def __getitem__(self, collectionKey):
        if collectionKey in self.collectionObjects:
            return self.collectionObjects[collectionKey]
        else:
            raise KeyError

    def __setitem__(self, collectionKey, item):
        self.collectionObjects[collectionKey] = item

    def __delitem__(self, collectionKey):
        del self.collectionObjects[collectionKey]

    def __iter__(self):
        return self.collectionObjects.__iter__()

    def __reversed__(self):
        ritems = Items()
        ritems.collectionObjects = self.collectionObjects.reversed()
        return ritems

    def __contains__(self, collectionKey):
        return collectionKey in self.collectionObjects

    def addCollection(self, collection):
        self.collectionObjects[collection.collectionKey] = collection
        self.orderedArray.append(collection)

    def getCollection(self, collectionKey):
        if collectionKey in self.collectionObjects:
            return self.collectionObjects[collectionKey]
        else:
            return False

    def addCollectionsFromFeed(self, feed):
        addedCollections = []
        for entry in feed.entries:
            collection = Collection(entry)
            self.addCollection(collection)
            addedCollections.append(collection)

        return addedCollections

    def writeCollection(self, collection):
        cols = self.writeCollections([collection])
        if cols == False:
            return False
        return cols[0]

    def writeCollections(self, collections):
        collectionsArray = []
        for collection in collections:
            collectionKey = collection.get('collectionKey')
            if collectionKey == '' or collectionKey == None:
                newCollectionKey = getKey()
                collection.set('collectionKey', newCollectionKey)
                collection.set('collectionVersion', 0)
            collectionsArray.append(collection)

        aparams = {'target': 'collections', 'content': 'json'}
        reqUrl = self.owningLibrary.apiRequestString(aparams)

        chunks = [collectionsArray[i:i + 50] for i in range(0, len(collectionsArray), 50)]
        for chunk in chunks:
            writeArray = []
            for collection in chunk:
                writeArray.append(collection.writeApiObject())
            requestData = json.dumps({'collections': writeArray})

            writeResponse = self.owningLibrary._request(reqUrl, 'POST', requestData, {'Content-Type': 'application/json'})
            if responseIsError(writeResponse):
                logging.info('writeCollections Error')
                logging.info(writeResponse.status_code)
                #entire request failed but we get no per-item write failure messages
                #so update all items with writeFailure manually
                for collection in chunk:
                    collection.writeFailure = {'code': writeResponse.status_code, 'message': writeResponse.text}
            else:
                updateObjectsFromWriteResponse(chunk, writeResponse)

        return collections

    def nestCollections(self):
        pass

    def orderCollections(self):
        pass

    def topCollectionKeys(self):
        pass



import sys
sys.path.append('../')
import unittest
import json
import logging
from libZotero import zotero


class TestItemFunctions(unittest.TestCase):

    def setUp(self):
        pass
        #self.seq = range(10)

    def test_templates(self):
        #zlib = zotero.Library('user', '475425', 'z_public_library', '')

        #create a new item of type book
        newItem = zotero.getTemplateItem('book')
        newItem.set('title', 'This is a book ---')
        newItem.set('abstractNote', 'Created using a zotero python library and the write api')

        self.assertEqual(newItem.get('title'), 'This is a book ---')
        self.assertEqual(newItem.get('abstractNote'), "Created using a zotero python library and the write api")

    def test_setandget(self):
        templateJsonString = """{"itemType":"journalArticle","title":"","creators":[{"creatorType":"author","firstName":"","lastName":""}],"abstractNote":"","publicationTitle":"","volume":"","issue":"","pages":"","date":"","series":"","seriesTitle":"","seriesText":"","journalAbbreviation":"","language":"","DOI":"","ISSN":"","shortTitle":"","url":"","accessDate":"","archive":"","archiveLocation":"","libraryCatalog":"","callNumber":"","rights":"","extra":"","tags":[],"collections":[],"relations":{}}"""

        templateArray = json.loads(templateJsonString)
        item = zotero.Item()
        item.initItemFromTemplate(templateArray)

        item.set('title', 'Journal Article Title')
        item.set('itemKey', 'ASDF1234')
        item.set('itemVersion', 74)
        item.set('itemType', 'conferencePaper')
        item.set('deleted', 1)
        item.set('parentItem', 'HJKL9876')
        item.set('abstractNote', 'This is a test item.')
        item.set('notRealField', 'Not a real field value.')

        #test that get returns what it should for each set
        self.assertEqual(item.get('title'), 'Journal Article Title', "get title should return what was set.")
        self.assertEqual(item.get('itemKey'), 'ASDF1234', "get itemkey should return what was set.")
        self.assertEqual(item.get('itemVersion'), 74, "get itemVersion should return what was set.")
        self.assertEqual(item.get('itemType'), 'conferencePaper', 'get itemType should return what was set.')
        self.assertEqual(item.get('deleted'), 1, "get deleted should return what was set.")
        self.assertEqual(item.get('parentItem'), 'HJKL9876', 'get parentItem should return what was set')
        self.assertEqual(item.get('abstractNote'), 'This is a test item.', 'get abstractNote should return what was set.')
        self.assertEqual(item.get('notRealField'), None, "get fake field value should return None.")

        self.assertEqual(item.title, 'Journal Article Title', 'title should be set on item object')
        self.assertEqual(item.apiObject['title'], 'Journal Article Title', 'title should be set on item apiObject')
        self.assertEqual(item.pristine['title'], "", 'title should not be set on item pristine')

        self.assertEqual(item.itemKey, 'ASDF1234', 'itemKey should be set on item object')
        self.assertEqual(item.apiObject['itemKey'], 'ASDF1234', 'itemKey should be set on item apiObject')
        self.assertNotIn('itemKey', item.pristine, 'isset(itemKey) should be false for pristine')

        self.assertEqual(item.itemVersion, 74, 'itemVersion should be set on item object')
        self.assertEqual(item.apiObject['itemVersion'], 74, 'itemVersion should be set on item apiObject')
        self.assertNotIn('itemVersion', item.pristine, 'isset(itemVersion) should be false for item pristine')

        self.assertEqual(item.apiObject['deleted'], 1, 'deleted should be set on item apiObject')

    def test_creators(self):
        templateJsonString = """{"itemType":"journalArticle","title":"","creators":[{"creatorType":"author","firstName":"","lastName":""}],"abstractNote":"","publicationTitle":"","volume":"","issue":"","pages":"","date":"","series":"","seriesTitle":"","seriesText":"","journalAbbreviation":"","language":"","DOI":"","ISSN":"","shortTitle":"","url":"","accessDate":"","archive":"","archiveLocation":"","libraryCatalog":"","callNumber":"","rights":"","extra":"","tags":[],"collections":[],"relations":{}}"""

        templateArray = json.loads(templateJsonString)
        item = zotero.Item()
        item.initItemFromTemplate(templateArray)

    def test_modifications(self):
        #log = logging.getLogger("TestItemFunctions.test_modifications")
        item = zotero.getTemplateItem('book')
        item.set('title', 'test book')
        item.addToCollection("ASDF1234")
        item.addToCollection("LKJH0987")

        itemCollections = item.get('collections')
        self.assertTrue('ASDF1234' in itemCollections)
        self.assertTrue('LKJH0987' in itemCollections)
        self.assertEqual(len(itemCollections), 2)

        itemTags = item.get('tags')
        self.assertEqual(len(itemTags), 0)

        item.addTag('Unread')
        item.addTag('purple')
        item.addTag('useless tripe')
        item.removeFromCollection('YUOI8765')  # remove from collection it doesn't belong to
        item.removeFromCollection('ASDF1234')

        itemCollections = item.get('collections')
        itemTags = item.get('tags')
        self.assertEqual(len(itemCollections), 1)
        self.assertEqual(len(itemTags), 3)
        self.assertIn('LKJH0987', itemCollections)
        self.assertNotIn('ASDF1234', itemCollections)
        self.assertIn('Unread', itemTags)
        self.assertIn('purple', itemTags)
        self.assertIn('useless tripe', itemTags)
        self.assertNotIn('Green', itemTags)

        item.removeTag('Unread')
        itemTags = item.get('tags')
        self.assertEqual(len(itemTags), 2)
        self.assertNotIn('Unread', itemTags)


if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger("TestItemFunctions.test_modifications").setLevel(logging.DEBUG)
    unittest.main()



import logging
import xml.dom.minidom
import re
import json
from copy import deepcopy
from entry import *
from zotero import responseIsError


class Item(Entry):
    fieldMap = {
        "creator"             : "Creator",
        "itemType"            : "Type",
        "title"               : "Title",
        "dateAdded"           : "Date Added",
        "dateModified"        : "Modified",
        "source"              : "Source",
        "notes"               : "Notes",
        "tags"                : "Tags",
        "attachments"         : "Attachments",
        "related"             : "Related",
        "url"                 : "URL",
        "rights"              : "Rights",
        "series"              : "Series",
        "volume"              : "Volume",
        "issue"               : "Issue",
        "edition"             : "Edition",
        "place"               : "Place",
        "publisher"           : "Publisher",
        "pages"               : "Pages",
        "ISBN"                : "ISBN",
        "publicationTitle"    : "Publication",
        "ISSN"                : "ISSN",
        "date"                : "Date",
        "section"             : "Section",
        "callNumber"          : "Call Number",
        "archiveLocation"     : "Loc. in Archive",
        "distributor"         : "Distributor",
        "extra"               : "Extra",
        "journalAbbreviation" : "Journal Abbr",
        "DOI"                 : "DOI",
        "accessDate"          : "Accessed",
        "seriesTitle"         : "Series Title",
        "seriesText"          : "Series Text",
        "seriesNumber"        : "Series Number",
        "institution"         : "Institution",
        "reportType"          : "Report Type",
        "code"                : "Code",
        "session"             : "Session",
        "legislativeBody"     : "Legislative Body",
        "history"             : "History",
        "reporter"            : "Reporter",
        "court"               : "Court",
        "numberOfVolumes"     : "# of Volumes",
        "committee"           : "Committee",
        "assignee"            : "Assignee",
        "patentNumber"        : "Patent Number",
        "priorityNumbers"     : "Priority Numbers",
        "issueDate"           : "Issue Date",
        "references"          : "References",
        "legalStatus"         : "Legal Status",
        "codeNumber"          : "Code Number",
        "artworkMedium"       : "Medium",
        "number"              : "Number",
        "artworkSize"         : "Artwork Size",
        "libraryCatalog"      : "Library Catalog",
        "videoRecordingType"  : "Recording Type",
        "interviewMedium"     : "Medium",
        "letterType"          : "Type",
        "manuscriptType"      : "Type",
        "mapType"             : "Type",
        "scale"               : "Scale",
        "thesisType"          : "Type",
        "websiteType"         : "Website Type",
        "audioRecordingType"  : "Recording Type",
        "label"               : "Label",
        "presentationType"    : "Type",
        "meetingName"         : "Meeting Name",
        "studio"              : "Studio",
        "runningTime"         : "Running Time",
        "network"             : "Network",
        "postType"            : "Post Type",
        "audioFileType"       : "File Type",
        "version"             : "Version",
        "system"              : "System",
        "company"             : "Company",
        "conferenceName"      : "Conference Name",
        "encyclopediaTitle"   : "Encyclopedia Title",
        "dictionaryTitle"     : "Dictionary Title",
        "language"            : "Language",
        "programmingLanguage" : "Language",
        "university"          : "University",
        "abstractNote"        : "Abstract",
        "websiteTitle"        : "Website Title",
        "reportNumber"        : "Report Number",
        "billNumber"          : "Bill Number",
        "codeVolume"          : "Code Volume",
        "codePages"           : "Code Pages",
        "dateDecided"         : "Date Decided",
        "reporterVolume"      : "Reporter Volume",
        "firstPage"           : "First Page",
        "documentNumber"      : "Document Number",
        "dateEnacted"         : "Date Enacted",
        "publicLawNumber"     : "Public Law Number",
        "country"             : "Country",
        "applicationNumber"   : "Application Number",
        "forumTitle"          : "Forum/Listserv Title",
        "episodeNumber"       : "Episode Number",
        "blogTitle"           : "Blog Title",
        "caseName"            : "Case Name",
        "nameOfAct"           : "Name of Act",
        "subject"             : "Subject",
        "proceedingsTitle"    : "Proceedings Title",
        "bookTitle"           : "Book Title",
        "shortTitle"          : "Short Title",
        "docketNumber"        : "Docket Number",
        "numPages"            : "# of Pages"
        }

    typeMap = {
        "note"                : "Note",
        "attachment"          : "Attachment",
        "book"                : "Book",
        "bookSection"         : "Book Section",
        "journalArticle"      : "Journal Article",
        "magazineArticle"     : "Magazine Article",
        "newspaperArticle"    : "Newspaper Article",
        "thesis"              : "Thesis",
        "letter"              : "Letter",
        "manuscript"          : "Manuscript",
        "interview"           : "Interview",
        "film"                : "Film",
        "artwork"             : "Artwork",
        "webpage"             : "Web Page",
        "report"              : "Report",
        "bill"                : "Bill",
        "case"                : "Case",
        "hearing"             : "Hearing",
        "patent"              : "Patent",
        "statute"             : "Statute",
        "email"               : "E-mail",
        "map"                 : "Map",
        "blogPost"            : "Blog Post",
        "instantMessage"      : "Instant Message",
        "forumPost"           : "Forum Post",
        "audioRecording"      : "Audio Recording",
        "presentation"        : "Presentation",
        "videoRecording"      : "Video Recording",
        "tvBroadcast"         : "TV Broadcast",
        "radioBroadcast"      : "Radio Broadcast",
        "podcast"             : "Podcast",
        "computerProgram"     : "Computer Program",
        "conferencePaper"     : "Conference Paper",
        "document"            : "Document",
        "encyclopediaArticle" : "Encyclopedia Article",
        "dictionaryEntry"     : "Dictionary Entry",
        }

    creatorMap = {
        "author"         : "Author",
        "contributor"    : "Contributor",
        "editor"         : "Editor",
        "translator"     : "Translator",
        "seriesEditor"   : "Series Editor",
        "interviewee"    : "Interview With",
        "interviewer"    : "Interviewer",
        "director"       : "Director",
        "scriptwriter"   : "Scriptwriter",
        "producer"       : "Producer",
        "castMember"     : "Cast Member",
        "sponsor"        : "Sponsor",
        "counsel"        : "Counsel",
        "inventor"       : "Inventor",
        "attorneyAgent"  : "Attorney/Agent",
        "recipient"      : "Recipient",
        "performer"      : "Performer",
        "composer"       : "Composer",
        "wordsBy"        : "Words By",
        "cartographer"   : "Cartographer",
        "programmer"     : "Programmer",
        "reviewedAuthor" : "Reviewed Author",
        "artist"         : "Artist",
        "commenter"      : "Commenter",
        "presenter"      : "Presenter",
        "guest"          : "Guest",
        "podcaster"      : "Podcaster"
        }

    def __init__(self, entryNode=None, library=None):
        self.itemKey = ''
        self.itemType = None
        self.creatorSummary = ''
        self.numChildren = 0
        self.numTags = 0
        self.year = ""
        self.childKeys = []
        self.parentKey = ''
        self.creators = []
        self.createdByUserID = None
        self.lastModifiedByUserID = None
        self.notes = []
        """
        @var int Represents the relationship of the child to the parent.
        0:file, 1:file, 2:snapshot, 3:web-link
        """
        self.linkMode = None
        self.mimeType = None
        self.parsedJson = None
        self.etag = ''
        """
        @var string content node of response useful if formatted bib
        request and we need to use the raw content
        """
        self.content = None
        self.bibContent = None
        self.subContents = {}
        self.apiObject = {'itemType': None, 'tags': [], 'collections': [], 'relations': []}
        self.pristine = {}
        self.owningLibrary = library
        if entryNode == None:
            return
        elif isinstance(entryNode, basestring):
            doc = xml.dom.minidom.parseString(entryNode)
            entryNode = doc.getElementsByTagName("entry").item(0)
        super(Item, self).__init__(entryNode)
        #check if we have multiple subcontent nodes
        subcontentNodes = entryNode.getElementsByTagNameNS("http://zotero.org/ns/api", "subcontent")

        #save raw Content node in case we need it
        if entryNode.getElementsByTagName("content").length > 0:
            d = entryNode.ownerDocument
            self.contentNode = entryNode.getElementsByTagName("content").item(0)
            self.content = d.toxml()
        # Extract the itemId and itemType
        self.itemKey = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', 'key').item(0).childNodes.item(0).nodeValue
        self.itemVersion = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', 'version').item(0).childNodes.item(0).nodeValue
        self.itemType = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', 'itemType').item(0).childNodes.item(0).nodeValue

        # Look for numTags node
        numTagsNode = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', "numTags").item(0)
        if numTagsNode:
            self.numTags = numTagsNode.childNodes.item(0).nodeValue

        # Look for year node
        yearNode = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', "year").item(0)
        if yearNode:
            if yearNode.childNodes.item(0):
                self.year = yearNode.childNodes.item(0).nodeValue

        # Look for numChildren node
        numChildrenNode = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', "numChildren").item(0)
        if numChildrenNode:
            self.numChildren = numChildrenNode.childNodes.item(0).nodeValue

        # Look for creatorSummary node
        creatorSummaryNode = entryNode.getElementsByTagNameNS('http://zotero.org/ns/api', "creatorSummary").item(0)
        if creatorSummaryNode:
            self.creatorSummary = creatorSummaryNode.childNodes.item(0).nodeValue

        #parse each subcontent node or the single content node
        if subcontentNodes.length > 0:
            for scnode in subcontentNodes:
                self.parseContentNode(scnode)
        else:
            contentNode = entryNode.getElementsByTagName('content').item(0)
            self.parseContentNode(contentNode)
        """
        if 'up' in self.links:
            parentLink = self.links['up']['href']
            matches = re.findall("/items\/([A-Z0-9]{8})/", parentLink)
            if len(matches):
                self.parentKey = matches[0]
        else:
            self.parentKey = False
        """

    def parseContentNode(self, contentNode):
        contentType = contentNode.getAttribute('type')
        zcontentType = contentNode.getAttributeNS('http://zotero.org/ns/api', 'type')

        if contentType == "" and zcontentType != "":
            contentType = zcontentType
        if (contentType == 'application/json') or (contentType == 'json'):
            self.pristine = json.loads(contentNode.childNodes.item(0).nodeValue)
            self.apiObject = json.loads(contentNode.childNodes.item(0).nodeValue)
            if 'creators' in self.apiObject:
                self.creators = self.apiObject['creators']
            else:
                self.creators = []

            self.itemVersion = self.apiObject['itemVersion'] if 'itemVersion' in self.apiObject else 0
            self.parentItemKey = self.apiObject['parentItem'] if 'parentItem' in self.apiObject else False

            if self.itemType == 'attachment':
                self.mimeType = self.apiObject['contentType']
                #TODO:translate mimetype
            if 'linkMode' in self.apiObject:
                self.linkMode = self.apiObject['linkMode']
            self.synced = True
        elif (contentType == 'bib'):
            bibNode = contentNode.getElementsByTagName('div').item(0)
            self.bibContent = bibNode.toxml()  # ownerDocument.saveXML(bibNode)

        contentString = ''
        childNodes = contentNode.childNodes
        for childNode in childNodes:
            contentString += childNode.toxml()
        self.subContents[contentType] = contentString

    def initItemFromTemplate(self, template):
        self.itemVersion = 0
        self.itemType = template['itemType']
        self.itemKey = ''
        self.pristine = deepcopy(template)
        self.apiObject = deepcopy(template)

    def get(self, key):
        if key == 'itemKey' or key == 'key':
            return self.itemKey
        elif key == 'itemVersion' or key == 'version':
            return self.itemVersion
        elif key == 'title':
            return self.title
        elif key == 'creatorSummary':
            return self.creatorSummary
        elif key == 'year':
            return self.year
        elif key == 'parentItem' or key == 'parentItemKey':
            return self.parentItemKey
        else:
            if key in self.apiObject:
                return self.apiObject[key]
            elif key in dir(self):
                return self.__getattribute__(key)
            else:
                return None

    def set(self, key, val):
        if key == "itemKey" or key == 'key':
            self.itemKey = val
            self.apiObject['itemKey'] = val
        elif key == "itemVersion" or key == 'version':
            self.itemVersion = val
            self.apiObject['itemVersion'] = val
        elif key == "title":
            self.title = val
        elif key == "itemType":
            self.itemType = val
            #TODO:translate fields
        elif key == "linkMode":
            #TODO:something here probably
            pass
        elif key == 'deleted':
            self.apiObject['deleted'] = val
        elif key == 'parentItem' or key == 'parentKey' or key == 'parentItemKey':
            if val == '':
                val = False
            self.parentItemKey = val
            self.apiObject['parentItem'] = val

        if key in self.apiObject:
            self.apiObject[key] = val
        if key in dir(self):
            self.__setattr__(key, val)

    def addCreator(self, creator):
        self.creators.append(creator)
        self.apiObject['creators'].append(creator)

    def writeApiObject(self):
        updateItem = dict(self.pristine.items() + self.apiObject.items())
        if 'creators' not in updateItem:
            return updateItem
        if len(updateItem['creators']) == 0:
            return updateItem

        newCreators = []
        for creator in updateItem['creators']:
            if creator.get('name', '') == '' and creator.get('firstName', '') == '' and creator.get('lastName', '') == '':
                continue
            else:
                newCreators.append(creator)
        updateItem['creators'] = newCreators
        return updateItem

    def updateItemObject(self):
        return self.writeApiObject()

    def newItemObject(self):
        newItem = self.apiObject
        newCreators = []
        if 'creators' in newItem:
            for creator in newItem['creators']:
                if 'creatorType' in creator:
                    if ('name' not in creator) and ('firstName' not in creator) and ('lastName' not in creator):
                        continue
                else:
                    newCreators.append(creator)
        newItem['creators'] = newCreators
        return newItem

    def addToCollection(self, collection):
        if isinstance(collection, basestring):
            collectionKey = collection
        else:
            collectionKey = collection.get('collectionKey')

        curCollections = self.get('collections')
        if collectionKey not in curCollections:
            curCollections.append(collectionKey)
        self.set('collections', curCollections)

    def removeFromCollection(self, collection):
        if isinstance(collection, basestring):
            collectionKey = collection
        else:
            collectionKey = collection.get('collectionKey')

        curCollections = self.get('collections')
        if collectionKey in curCollections:
            curCollections.remove(collectionKey)
            self.set('collections', curCollections)

    def addTag(self, newtagname, type=None):
        itemTags = self.get('tags')
        #assumes we'll get an array
        for tag in itemTags:
            if isinstance(tag, basestring) and tag == newtagname:
                return
            elif 'tag' in tag and tag['tag'] == newtagname:
                return

        if type != None:
            itemTags.append({'tag': newtagname, 'type': type})
        else:
            itemTags.append(newtagname)
        self.set('tags', itemTags)

    def removeTag(self, rmtagname):
        itemTags = self.get('tags')
        #assumes we'll get an array
        for tag in itemTags[:]:
            if (isinstance(tag, basestring) and tag == rmtagname) or (isinstance(tag, dict) and 'tag' in tag and tag['tag'] == rmtagname):
                del itemTags[itemTags.index(tag)]
                self.set('tags', itemTags)
                return

    def addNote(self, note):
        self.notes.append(note)
        parentItemKey = self.get('itemKey')
        if parentItemKey != None:
            note.set('parentItem', parentItemKey)
        return

    def trashItem(self):
        self.set('deleted', 1)

    def untrashItem(self):
        self.set('deleted', 0)

    def save(self):
        self.owningLibrary.items.writeItems([self])

    def getChildren(self):
        #short circuit if has item has no children
        if self.numChildren == 0 or self.parentItemKey != false:
            return []

        config = {'target': 'children', 'libraryType': self.owningLibrary.libraryType, 'libraryID': self.owningLibrary.libraryID, 'itemKey': self.itemKey, 'content': 'json'}
        requestUrl = self.owningLibrary.apiRequestString(config)

        response = self.owningLibrary._request(requestUrl, 'GET')

        #load response into item objects
        fetchedItems = []
        if responseIsError(response):
            return False

        feed = zotero.Feed(response.text)
        fetchedItems = self.owningLibrary.items.addItemsFromFeed(feed)
        return fetchedItems

    def compareItem(self, item):
        pass



import logging
from item import *
from zotero import getKey, responseIsError, updateObjectsFromWriteResponse


class Items(object):
    def __init__(self):
        self.itemObjects = {}
        self.itemsVersion = 0
        self.owningLibrary = None

    def __len__(self):
        return len(self.itemObjects)

    def __getitem__(self, itemKey):
        if itemKey in self.itemObjects:
            return self.itemObjects[itemKey]
        else:
            raise KeyError

    def __setitem__(self, itemKey, item):
        self.itemObjects[itemKey] = item

    def __delitem__(self, itemKey):
        del self.itemObjects[itemKey]

    def __iter__(self):
        return self.itemObjects.__iter__()

    def __reversed__(self):
        ritems = Items()
        ritems.itemObjects = self.itemObjects.reversed()
        return ritems

    def __contains__(self, itemKey):
        return itemKey in self.itemObjects

    def getItem(self, itemKey):
        if itemKey in self.itemObjects:
            return self.itemObjects[itemKey]
        else:
            return None

    def addItem(self, item):
        itemKey = item.itemKey
        self.itemObjects[itemKey] = item
        return item

    def addItemsFromFeed(self, feed):
        addedItems = []
        for entry in feed.entries:
            item = Item(entry)
            self.addItem(item)
            addedItems.append(item)
        return addedItems

    def replaceItem(self, item):
        itemKey = item.itemKey
        self.itemObjects[itemKey] = item

    def writeItem(self, item):
        written = self.writeItems([item])
        return written[0]

    def writeItems(self, items):
        writeItems = []

        for item in items:
            itemKey = item.get('itemKey')
            if itemKey == '':
                newItemKey = getKey()
                item.set('itemKey', newItemKey)
                item.set('itemVersion', 0)
            writeItems.append(item)

            #add separate note items if this item has any
            itemNotes = item.get('notes')
            if itemNotes and len(itemNotes) > 0:
                for note in itemNotes:
                    note.set('parentItem', item.get('itemKey'))
                    note.set('itemKey', getKey())
                    note.set('itemVersion', 0)
                    writeItems.append(note)

        aparams = {'target': 'items', 'content': 'json'}
        reqUrl = self.owningLibrary.apiRequestString(aparams)

        chunks = [writeItems[i:i + 50] for i in range(0, len(writeItems), 50)]
        for chunk in chunks:
            writeArray = []
            for item in chunk:
                writeArray.append(item.writeApiObject())
            requestData = json.dumps({'items': writeArray})

            writeResponse = self.owningLibrary._request(reqUrl, 'POST', requestData, {'Content-Type': 'application/json'})
            if responseIsError(writeResponse):
                logging.info('writeItems Error')
                logging.info(writeResponse.status_code)
                #entire request failed but we get no per-item write failure messages
                #so update all items with writeFailure manually
                for item in chunk:
                    item.writeFailure = {'code': writeResponse.status_code, 'message': writeResponse.text}
            else:
                updateObjectsFromWriteResponse(chunk, writeResponse)
        return writeItems

    def deleteItem(self, item):
        """Permanently delete an existing item."""
        aparams = {'target': 'item', 'itemKey': item.itemKey}
        reqUrl = self.owningLibrary.apiRequestString(aparams)
        response = self.owningLibrary_request(reqUrl, 'DELETE', None, {'If-Unmodified-Since-Version': item.get('itemVersion')})
        return response

    def deleteItems(self, items, version=None):
        """Permanently delete up to 50 existing items."""
        if len(items) > 50:
            raise "Cannot delete more than 50 items at once."
        itemKeys = []
        latestItemVersion = 0
        for item in items:
            itemKeys.append(item.get('itemKey'))
            v = item.get('itemVersion')
            if v > latestItemVersion:
                latestItemVersion = v

        if version == None:
            if self.itemsVersion != 0:
                version = self.itemsVersion
            else:
                version = latestItemVersion

        itemKeyString = ",".join(itemKeys)
        aparams = {'target': 'items', 'itemKey': itemKeyString}
        reqUrl = self.owningLibrary.apiRequestString(aparams)
        response = self.owningLibrary._request(reqUrl, 'DELETE', None, {'If-Unmodified-Since-Version': version})
        return response

    def trashItem(self, item):
        item.trashItem()
        return item.save()

    def trashItems(self, items):
        for item in items:
            item.trashItem()
        return self.writeItems(items)



import urllib
import urllib2
import urlparse
#import requests
import json
import xml.dom.minidom
from random import randint
import logging
import pickle
import zotero


def randomString(length=0, chars=None):
    if chars == None:
        chars = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    if length == 0:
        length = 8
    randomstring = ''
    for i in range(length):
        rnum = randint(0, len(chars) - 1)
        randomstring += chars[rnum]
    return randomstring


def getKey():
    baseString = "23456789ABCDEFGHIJKMNPQRSTUVWXZ"
    return randomString(8, baseString)


def updateObjectsFromWriteResponse(objsArray, response):
    data = json.loads(response.text)
    if(response.status_code == 200):
        newLastModifiedVersion = response.headers["Last-Modified-Version"]
        if 'success' in data:
            for ind in data['success'].keys():
                key = data['success'][ind]
                i = int(ind)
                obj = objsArray[i]

                objKey = obj.get('key')
                if objKey != '' and objKey != key:
                    raise zotero.ZoteroApiError("Item key mismatch in multi-write request")
                if objKey == '':
                    obj.set('key', key)
                obj.set('version', newLastModifiedVersion)
                obj.synced = True
                obj.writeFailure = False

        if 'failed' in data:
            for ind in data['failed'].keys():
                val = data['failed'][ind]
                i = int(ind)
                obj = objsArray[i]
                obj.writeFailure = val
    elif response.status_code == 204:
        objsArray[0].synced = True


def apiRequestUrl(params={}, base=None):
    """Build and return a valid url to request a Zotero API resource."""
    if base == None:
        base = zotero.ZOTERO_URI
    if 'target' not in params:
        raise zotero.ZoteroUrlError("No target defined for api request")
    if 'libraryType' not in params and params['target'] != 'itemTemplate':
        raise zotero.ZoteroUrlError("No libraryType defined for api request")
    if 'libraryID' not in params and params['target'] != 'itemTemplate':
        raise zotero.ZoteroUrlError("No libraryID defined for api request")
    #special elif for www based api requests until those methods are mapped for api.zotero
    if params['target'] == 'user' or params['target'] == 'cv':
        base = 'https://www.zotero.org/api'
    elif params['target'] == 'itemTemplate':
        return base + '/items/new'

    if (params['libraryType'] == 'user') or (params['libraryType'] == 'group'):
        url = base + '/' + params['libraryType'] + 's/' + str(params['libraryID'])
    else:
        url = base

    if ('collectionKey' in params) and (params['collectionKey']):
        if params['collectionKey'] == 'trash':
            url += '/items/trash'
            return url
        else:
            url += '/collections/' + params['collectionKey']

    if params['target'] == 'items':
        url += '/items'
    elif params['target'] == 'item':
        if 'itemKey' in params:
            url += '/items/' + params['itemKey']
        else:
            url += '/items'
    elif params['target'] == 'collections':
        url += '/collections'
    elif params['target'] == 'collection':
        pass  # already added above because collectionKey would have to be set
    elif params['target'] == 'tags':
        url += '/tags'
    elif params['target'] == 'children':
        url += '/items/' + params['itemKey'] + '/children'
    elif params['target'] == 'itemTemplate':
        url = base + '/items/new'
    elif params['target'] == 'key':
        url = base + '/users/' + str(params['userID']) + '/keys/' + params['apiKey']
    elif params['target'] == 'userGroups':
        url = base + '/users/' + str(params['userID']) + '/groups'
    elif params['target'] == 'trash':
        url += '/items/trash'
    elif params['target'] == 'cv':
        url += '/cv'
    elif params['target'] == 'deleted':
        url += '/deleted'
    if 'targetModifier' in params:
        if params['targetModifier'] == 'top':
            url += '/top'
        elif params['targetModifier'] == 'children':
            url += '/children'
        elif params['targetModifier'] == 'file':
            if params['target'] != 'item':
                raise zotero.ZoteroUrlError('Trying to get file on non-item target')
            url += '/file'
        elif params['targetModifier'] == 'fileview':
            if params['target'] != 'item':
                raise zotero.ZoteroUrlError('Trying to get file on non-item target')
            url += '/file/view'
    #print "apiRequestUrl: " . url . "\n"
    return url


def apiQueryString(passedParams={}):
    """Build and return a valid query string for a request for a Zotero API resource."""
    # Tags query formats
    #
    # ?tag=foo
    # ?tag=foo bar # phrase
    # ?tag=-foo # negation
    # ?tag=\-foo # literal hyphen (only for first character)
    # ?tag=foo&tag=bar # AND
    # ?tag=foo&tagType=0
    # ?tag=foo bar || bar&tagType=0
    queryParamOptions = ['start',
                         'limit',
                         'order',
                         'sort',
                         'content',
                         'q',
                         'itemType',
                         'locale',
                         'key',
                         'itemKey',
                         'tag',
                         'tagType',
                         'style',
                         'format',
                         'linkMode',
                         'upload',
                         'algorithm']
    #build simple api query parameters object
    logging.info(passedParams)
    queryParams = []
    for val in sorted(queryParamOptions):
        if (val in passedParams) and (passedParams[val] != ''):
            if (val == 'itemKey') and ('target' in passedParams) and (passedParams['target'] != 'items'):
                #itemKey belongs in url, not querystring
                pass
            else:
                if val == 'itemKey' and not isinstance(val, basestring):
                    queryParams.append((val, ",".join(passedParams[val])))
                else:
                    queryParams.append((val, passedParams[val]))
    #print(queryParams)
    return '?' + urllib.urlencode(queryParams)


def apiRequestString(params):
    return apiRequestUrl(params) + apiQueryString(params)


def zrequest(url, method='GET', body=None, headers={}):
    """Make a request to the Zotero API and return the response object."""
    opener = urllib2.build_opener(urllib2.HTTPHandler)
    req = urllib2.Request(url, body)
    for key, val in headers.items():
        req.add_header(key, val)
    #add Zotero api version header if not explicitly set by arguments
    if 'Zotero-API-Version' not in headers.items():
        req.add_header('Zotero-API-Version', 2)
    req.get_method = lambda: method
    r = None
    try:
        res = opener.open(req)
        r = zotero.Response(res)
        return r
    except (urllib2.URLError, urllib2.HTTPError) as err:
        r = zotero.Response(err)
        return r


def responseIsError(response):
    """Check if a response object is an error or not"""
    return response.status_code >= 400


def getTemplateItem(itemType, linkMode=None):
    """Return a template for a Zotero API item of a particular type."""
    newItem = zotero.Item()
    aparams = {'target': 'itemTemplate', 'itemType': itemType}
    if linkMode != None:
        aparams['linkMode'] = linkMode

    reqUrl = apiRequestString(aparams)
    response = zrequest(reqUrl)
    if response.status_code != 200:
        raise zotero.ZoteroApiError("Error getting template item")
    itemTemplate = json.loads(response.text)
    newItem.initItemFromTemplate(itemTemplate)
    return newItem


def loadLibrary(picklestring):
    return pickle.loads(picklestring)


class Library(object):
    def __init__(self,
                 libraryType,
                 libraryID,
                 libraryUrlIdentifier,
                 apiKey,
                 baseWebsiteUrl="http://www.zotero.org",
                 cachettl=300):
        self.ZOTERO_URI = 'https://api.zotero.org'
        self._apiKey = apiKey
        self._followRedirects = True
        self.libraryType = libraryType
        self.libraryID = libraryID
        self.libraryString = self.libraryString(libraryType, libraryID)
        self.libraryUrlIdentifier = libraryUrlIdentifier
        self.libraryBaseWebsiteUrl = baseWebsiteUrl
        self.items = zotero.Items()
        self.items.owningLibrary = self
        self.collections = zotero.Collections()
        self.collections.owningLibrary = self
        self.dirty = False
        self.useLibraryAsContainer = True
        self._lastResponse = None
        self._lastFeed = None
        self._cachettl = cachettl
        self._cachePrefix = 'libZotero'
        self._cache = zotero.Cache()
        if cachettl > 0:
            self._cacheResponses = True
        else:
            self._cacheResponses = False

    def libraryString(self, ltype, libraryID):
        """Return a string that uniquely identifies a library for use as a cache key."""
        lstring = ''
        if(ltype == 'user'):
            lstring = 'u'
        elif(ltype == 'group'):
            lstring = 'g'
        lstring += str(libraryID)
        return lstring

    def apiRequestUrl(self, params={}, base=None):
        """Build and return a valid url to request a Zotero API resource."""
        if 'target' not in params:
            raise zotero.ZoteroUrlError("No target defined for api request")
        #fill library specific params in if not present
        if 'libraryType' not in params:
            params['libraryType'] = self.libraryType
        if 'libraryID' not in params:
            params['libraryID'] = self.libraryID

        return apiRequestUrl(params)

    def apiQueryString(self, passedParams={}):
        """Build and return a valid query string for a request for a Zotero API resource."""
        if 'key' not in passedParams:
            passedParams['key'] = self._apiKey
        return apiQueryString(passedParams)

    def apiRequestString(self, params={}):
        return self.apiRequestUrl(params) + self.apiQueryString(params)

    def _request(self, url, method='GET', body=None, headers={}):
        """Make a request to the Zotero API and return the response object."""
        logging.debug("zotero.Library._request")
        #check for cached result before http request
        r = None
        logging.debug("checking for cached request")
        if (self._cacheResponses) and (method.upper() == 'GET'):
            logging.debug("caching responses and get request - checking cache")
            r = self._cache.cache_fetch(url)
        if(r != None):
            self._lastResponse = r
            return r
        r = zrequest(url, method, body, headers)
        if self._cacheResponses and (method.upper() == 'GET'):
            self._cache.cache_store(url, r)
        self._lastResponse = r
        return r

    def fetchCollections(self, params={}):
        """Fetch a set of collections."""
        aparams = {'target': 'collections', 'content': 'json', 'limit': 100}
        aparams.update(params)
        reqUrl = self.apiRequestString(aparams)
        response = self._request(reqUrl)
        if response.status_code != 200:
            raise zotero.ZoteroApiError("Error fetching collections")
        feed = zotero.Feed(response.text)
        self._lastFeed = feed
        addedCollections = self.collections.addCollectionsFromFeed(feed)

        if 'next' in feed.links:
            nextUrl = feed.links['next']['href']
            parsedNextUrl = urlparse.urlparse(nextUrl)
            parsedNextQuery = urlparse.parse_qs(parsedNextUrl.query)
            parsedNextQuery = self.apiQueryString(parsedNextQuery.update({'key': self._apiKey}))
            reqUrl = parsedNextUrl['scheme'] + '://' + parsedNextUrl['host'] + parsedNextUrl['path'] + parsedNextQuery
        else:
            reqUrl = False
        return addedCollections

    def fetchItemsTop(self, params={}):
        """Fetch a set of top level items."""
        params['targetModifier'] = 'top'
        return self.fetchItems(params)

    def fetchItems(self, params={}):
        """Fetch a set of items."""
        fetchedItems = []
        aparams = {'target': 'items', 'content': 'json', 'key': self._apiKey}
        aparams.update(params)
        reqUrl = self.apiRequestString(aparams)
        logging.info(reqUrl)
        response = self._request(reqUrl)
        if(response.status_code != 200):
            raise zotero.ZoteroApiError("Error fetching items. " + str(response.status_code))
        body = response.text
        feed = zotero.Feed(body)
        self._lastFeed = feed

        fetchedItems = self.items.addItemsFromFeed(feed)

        return fetchedItems

    def fetchItemKeys(self, params={}):
        """Fetch all item keys in the library, specified by params."""
        logging.info('zotero.Library.fetchItemKeys')
        fetchedKeys = []
        aparams = {'target': 'items', 'format': 'keys'}
        aparams.update(params)
        reqUrl = self.apiRequestString(aparams)
        response = self._request(reqUrl)
        if response.status_code != 200:
            raise zotero.ZoteroApiError("Error fetching item keys" + str(response.status_code))
        body = response.text
        fetchedKeys = body.strip().split("\n")
        return fetchedKeys

    def fetchTrashedItems(self, params={}):
        """Fetch a set of items marked for deletion."""
        fetchedItems = []
        aparams = {'target': 'trash', 'content': 'json'}
        aparams.update(params)
        reqUrl = self.apiRequestString(aparams)
        response = self._request(reqUrl)
        if response.status_code != 200:
            raise zotero.ZoteroApiError("Error fetching items" + str(response.status_code))
        body = response.text
        feed = zotero.Feed(body)
        self._lastFeed = feed
        fetchedItems = self.items.addItemsFromFeed(feed)
        return fetchedItems

    def fetchItemsAfter(self, itemKey, params={}):
        """Fetch a set of items after the specified itemKey."""
        #this might be completely broken
        fetchedItems = []
        itemKeys = self.fetchItemKeys(params)
        if itemKey != '':
            if itemKey not in itemKeys:
                return fetchedItems

        index = itemKeys.index(itemKey)
        offset = 0
        while (offset < index):
            if index - offset > 50:
                uindex = offset + 50
            else:
                uindex = index
            itemKeysToFetch = itemKeys[0:uindex]
            offset = uindex
            params['itemKey'] = itemKeysToFetch.join(',')
            fetchedSet = self.fetchItems(params)
            fetchedItems = fetchedItems.concat(fetchedSet)

        return fetchedItems

    def fetchItem(self, itemKey, params={}):
        """Fetch a single item."""
        aparams = {'target': 'item', 'content': 'json', 'itemKey': itemKey}
        aparams.update(params)
        reqUrl = self.apiRequestString(aparams)

        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            raise zotero.ZoteroApiError("Error fetching items")

        body = response.text
        item = zotero.Item(body)
        if not item:
            return False
        self.items.addItem(item)
        return item

    def fetchItemBib(self, itemKey, style=None):
        """Fetch a bibliography entry for an item."""
        #TODO:parse correctly and return just bib
        aparams = {'target': 'item', 'content': 'bib', 'itemKey': itemKey}
        if style != None:
            aparams['style'] = style
        reqUrl = self.apiRequestString(aparams)

        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            raise zotero.ZoteroApiError("Error fetching items")
        body = response.text
        feed = zotero.Feed(body)
        if len(feed.entries) == 0:
            return False
        else:
            item = zotero.Item(feed.entries[0])
            self.items.addItem(item)
            return item

    def itemDownloadLink(self, itemKey):
        """Get the link to download an attached item file."""
        aparams = {'target': 'item', 'itemKey': itemKey, 'targetModifier': 'file'}
        return self.apiRequestString(aparams)

    def writeUpdatedItem(self, item):
        """Attempt to write a modified item back to the server."""
        if item.owningLibrary == None:
            item.associateWithLibrary(self)
        return self.items.writeItem(item)

    def uploadNewAttachedFile(self, item, filedata, fileinfo):
        """Create an attachment item as a child of the passed item and upload
        a file as the attachment.
        """
        #get upload authorization
        #post file or patch
        uaparams = {'target': 'item', 'targetModifier': 'file', 'itemKey': item.itemKey}
        reqUrl = self.apiRequestString(uaparams)
        uaPostData = urllib.urlencode(fileinfo)
        uploadAuthResponse = zrequest(reqUrl, 'POST', uaPostData, {'If-None-Match': '*'})
        if uploadAuthResponse.status_code != 200:
            logging.info('upload new attached file - uploadAuthResponse: ')
            logging.info(uploadAuthResponse.status_code)
            logging.info(uploadAuthResponse.text)
            raise zotero.ZoteroApiError("Upload Authorization Failed")
        #full upload
        upAuthOb = json.loads(uploadAuthResponse.text)
        if 'exists' in upAuthOb and upAuthOb['exists'] == 1:
            #file already exists with this hash
            return None
        #uploadBody = u'' + upAuthOb['prefix'] + filedata + upAuthOb['suffix']
        uploadBody = bytearray(upAuthOb['prefix'].encode())
        uploadBody.extend(filedata)
        uploadBody.extend(bytearray(upAuthOb['suffix'].encode()))

        uploadResponse = zrequest(upAuthOb['url'], 'POST', uploadBody, {'Content-Type': upAuthOb['contentType']})
        if uploadResponse.status_code != 201:
            raise zotero.ZoteroApiError("Error uploading attachment file")
        ucparams = {'target': 'item', 'targetModifier': 'file', 'itemKey': item.itemKey}
        ucReqUrl = self.apiRequestString(ucparams)
        registerUploadBody = uaPostData = urllib.urlencode({'upload': upAuthOb['uploadKey']})
        ucResponse = zrequest(ucReqUrl, 'POST', registerUploadBody, {'Content-Type': 'application/x-www-form-urlencoded',
                                                                     'If-None-Match': '*'})
        if ucResponse.status_code != 204:
            raise zotero.ZoteroApiError("Error confirming upload to Zotero API - " + ucResponse.text)
        return True

    def uploadAttachedFilePatch(self, item, patchdata, fileinfo, algorithm='bsdiff'):
        """Upload a patch for an attached file already present on the server."""
        #get upload authorization
        #post file or patch
        uaparams = {'target': 'item', 'targetModifier': 'file', 'itemKey': item.itemKey}
        reqUrl = self.apiRequestString(uaparams)
        uaPostData = urllib.urlencode(fileinfo)
        uploadAuthResponse = zrequest(reqUrl, 'POST', uaPostData, {'If-Match': item.get('md5')})
        if uploadAuthResponse.status_code != 200:
            logging.info('upload new attached file - uploadAuthResponse: ')
            logging.info(uploadAuthResponse.status_code)
            logging.info(uploadAuthResponse.text)
            raise zotero.ZoteroApiError("Upload Authorization Failed")
        #patch upload
        upAuthOb = json.loads(uploadAuthResponse.text)
        if 'exists' in upAuthOb and upAuthOb['exists'] == 1:
            #file already exists with this hash
            return None
        upparams = {'target': 'item', 'targetModifier': 'file', 'itemKey': item.itemKey, 'upload': upAuthOb['uploadKey'], 'algorithm': algorithm}
        uploadUrl = self.apiRequestString(upparams)
        logging.info(upAuthOb)
        #uploadBody = bytearray(upAuthOb['prefix'].encode())
        #uploadBody.extend(patchdata)
        #uploadBody.extend(bytearray(upAuthOb['suffix'].encode()))
        uploadBody = bytearray(patchdata)
        logging.info(uploadBody)
        logging.info(upAuthOb['contentType'])
        uploadResponse = zrequest(uploadUrl, 'PATCH', uploadBody, {'Content-Type': upAuthOb['contentType'],
                                                                   'If-Match': item.get('md5')})
        if uploadResponse.status_code != 204:
            logging.info(uploadResponse.status_code)
            logging.info(uploadResponse.text)
            raise zotero.ZoteroApiError("Error uploading or applying attachment file patch")
        return True

    def uploadExistingAttachedFile(self, item, f, fileinfo):
        """Upload a full copy of a file for an attachment that already exists on the server."""
        pass

    def createAttachmentItem(self, parentItem, attachmentInfo):
        """Create a new attachment item as a child of parentItem."""
        logging.info("createAttachmentItem")
        #get attachment template
        adata = {'attachmentType': 'imported_file', 'contentType': None, 'filename': ''}
        adata.update(attachmentInfo)
        logging.info("createAttachmentItem: " + str(adata))
        templateItem = self.getTemplateItem('attachment', adata['attachmentType'])
        templateItem.set('parentItem', parentItem.itemKey)
        templateItem.set('title', adata['filename'])
        templateItem.set('contentType', adata['contentType'])
        #create child item
        logging.info("creating attachment Item: ")
        return self.createItem(templateItem)

    def getTemplateItem(self, itemType, linkMode=None):
        """Return a template for a Zotero API item of a particular type."""
        newItem = zotero.Item()
        aparams = {'target': 'itemTemplate', 'itemType': itemType}
        if linkMode != None:
            aparams['linkMode'] = linkMode

        reqUrl = self.apiRequestString(aparams)
        response = self._request(reqUrl)
        if response.status_code != 200:
            raise zotero.ZoteroApiError("Error getting template item")
        itemTemplate = json.loads(response.text)
        newItem.initItemFromTemplate(itemTemplate)
        newItem.owningLibrary = self
        return newItem

    def createItem(self, item):
        """Create a new item on the server."""
        logging.info("createItem")
        writtenItems = self.items.writeItems([item])
        if writtenItems != False:
            return writtenItems[0]
        return False

    def addNotes(self, parentItem, noteItem):
        """Add note items as children of parentItem."""
        logging.info(noteItem)
        noteItem.set('parentItem', parentItem.get('itemKey'))
        writtenItems = self.items.writeItems([noteItem])
        return writtenItems[0]

    def createCollection(self, name, parent=None):
        """Create a new collection on the server."""
        collection = zotero.Collection(library=self)
        collection.set('name', name)
        collection.set('parentCollectionKey', parent)
        return self.collections.writeCollection(collection)

    def removeCollection(self, collection):
        """Remove an existing collection from the server."""
        aparams = {'target': 'collection', 'collectionKey': collection.collectionKey}
        reqUrl = self.apiRequestString(aparams)
        response = self._request(reqUrl, 'DELETE', None, {'If-Unmodified-Since-Version': collection.get('collectionVersion')})
        return response

    def addItemsToCollection(self, collection, items):
        """Add specified items to the specified collection."""
        for item in items:
            item.addToCollection(collection)
        updatedItems = self.items.writeItems(items)
        return updatedItems

    def removeItemFromCollection(self, collection, item):
        """Remove item from collection."""
        item.removeFromCollection(collection)
        return self.items.writeItems([item])

    def removeItemsFromCollection(self, collection, items):
        """Remove multiple items from the specified collection."""
        for item in items:
            item.removeFromCollection(collection)
        return self.items.writeItems(items)

    def writeUpdatedCollection(self, collection):
        """Submit a modified collection to be saved on the server."""
        return self.collections.writeCollection(collection)

    def deleteItem(self, item):
        """Permanently delete an existing item."""
        return self.items.deleteItem(item)

    def trashItem(self, item):
        """Mark an existing item for deletion, adding it to the trash metacollection."""
        item.trashItem()
        return self.items.writeItems([item])

    def fetchItemChildren(self, item):
        """Fetch child items of the specified item."""
        return item.getChildren()

    def getItemTypes(self):
        """Get the list of possible Zotero item types."""
        reqUrl = zotero.ZOTERO_URI + 'itemTypes'
        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            raise zotero.ZoteroApiError("failed to fetch itemTypes")
        itemTypes = json.loads(response.getBody())
        return itemTypes

    def getItemFields(self, itemType):
        """Get the list of possible item fields for a particular Zotero item type."""
        reqUrl = zotero.ZOTERO_URI + 'itemFields'
        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            raise zotero.ZoteroApiError("failed to fetch itemFields")
        itemFields = json.loads(response.getBody())
        return itemFields

    def getCreatorTypes(self, itemType):
        """Get the list of possible creator types for a particular Zotero item type."""
        reqUrl = zotero.ZOTERO_URI + 'itemTypeCreatorTypes?itemType=' + itemType
        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            raise zotero.ZoteroApiError("failed to fetch creatorTypes")
        creatorTypes = json.loads(response.getBody())
        return creatorTypes

    def getCreatorFields(self, creatorType):
        """Get the list of creator fields and translations for a particular creator type."""
        reqUrl = zotero.ZOTERO_URI + 'creatorFields'
        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            raise zotero.ZoteroApiError("failed to fetch creatorFields")
        creatorFields = json.loads(response.getBody())
        return creatorFields

    def fetchAllTags(self, params):
        """Fetch all tags, even over multiple requests, present in the library."""
        aparams = {'target': 'tags', 'content': 'json', 'limit': 50}
        aparams.update(params)
        reqUrl = self.apiRequestString(aparams)
        while True:
            response = self._request(reqUrl, 'GET')
            if response.status_code != 200:
                return False
            feed = zotero.Feed(response.text)
            entries = feed.entries
            tags = []
            for entry in entries:
                tag = zotero.Tag(entry)
                tags.append(tag)

            if 'next' in feed.links:
                nextUrl = feed.links['next']['href']
                #Add the apikey to the reqUrl querystring since feeds leave it out
                parsedNextUrl = urlparse.urlparse(nextUrl)
                parsedNextQuery = urlparse.parse_qs(parsedNextUrl.query)
                parsedNextQuery = self.apiQueryString(parsedNextQuery.update({'key': self._apiKey}))
                reqUrl = parsedNextUrl['scheme'] + '://' + parsedNextUrl['host'] + parsedNextUrl['path'] + parsedNextQuery
            else:
                break
        return tags

    def fetchTags(self, params):
        """Make a single request to get a set of tags."""
        aparams = {'target': 'tags', 'content': 'json', 'limit': 50}
        aparams.update(params)
        reqUrl = self.apiRequestString(aparams)

        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            return False
        feed = zotero.Feed(response.text)
        entries = feed.entries
        tags = []
        for entry in entries:
            tag = zotero.Tag(entry)
            tags.append(tag)
        return tags

    def getKeyPermissions(self, userID, key):
        """Get information about the permissions a particular key has."""
        if userID == None:
            userID = self.libraryID
        if key == False:
            if self._apiKey == '':
                return False
            key = self._apiKey

        reqUrl = self.apiRequestString({'target': 'key', 'apiKey': key, 'userID': userID})
        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            return False
        doc = xml.dom.minidom.parseString(response.text)
        keyNode = doc.getElementsByTagName('key').item(0)
        keyPerms = self.parseKey(keyNode)
        return keyPerms

    def parseKey(self, keyNode):
        """Parse the api key xml returned by the Zotero API."""
        keyPerms = {"library": "0", "notes": "0", "write": "0", 'groups': {}}

        accessEls = keyNode.getElementsByTagName('access')
        for access in accessEls:
            libraryAccess = access.getAttribute("library")
            notesAccess = access.getAttribute("notes")
            writeAccess = access.getAttribute("write")
            groupAccess = access.getAttribute("group")
            groupPermission = 'write' if access.getAttribute("write") else 'read'

            if libraryAccess:
                keyPerms['library'] = libraryAccess
            if notesAccess:
                keyPerms['notes'] = notesAccess
            if groupAccess:
                keyPerms['groups'][groupAccess] = groupPermission
            elif writeAccess:
                keyPerms['write'] = writeAccess
        return keyPerms

    def fetchGroups(self, userID):
        """Fetch the set of groups a user is a member of."""
        if not userID:
            userID = self.libraryID
        aparams = {'target': 'userGroups', 'userID': userID, 'content': 'json'}
        reqUrl = self.apiRequestString(aparams)
        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            return False

        doc = xml.dom.minidom.parseString(response.text)
        entries = doc.getElementsByTagName('entry')
        groups = []
        for entry in entries:
            group = zotero.Group(entry)
            groups.append(group)
        return groups

    def getCV(self, userID):
        """Get a user's C.V."""
        if userID == '' and self.libraryType == 'user':
            userID = self.libraryID
        aparams = {'target': 'cv', 'libraryType': 'user', 'libraryID': userID}
        reqUrl = self.apiRequestString(aparams)

        response = self._request(reqUrl, 'GET')
        if response.status_code != 200:
            return False

        doc = xml.dom.minidom.parseString(response.text)
        sectionNodes = doc.getElementsByTagNameNS('*', 'cvsection')
        sections = []
        for sectionNode in sectionNodes:
            sectionTitle = sectionNode.getAttribute('title')
            c = sectionNode.nodeValue
            sections.append({'title':  sectionTitle, 'content': c})
        return sections

    def saveLibrary(self):
        """Return pickled self for storage"""
        return pickle.dumps(self)

if __name__ == "__main__":
    pass



import logging
from exceptions import *
import xml.dom.minidom


class Entry(object):
    def __init__(self, entryNode):
        if isinstance(entryNode, basestring):
            logging.info("doc is a string. make it a dom")
            entryNode = xml.dom.minidom.parseString(entryNode)
            if not entryNode:
                raise ZoteroParseError("Error constructing feed Domdoc")

        self.title = entryNode.getElementsByTagName('title').item(0).childNodes.item(0).nodeValue
        self.id = entryNode.getElementsByTagName('id').item(0).childNodes.item(0).nodeValue
        self.dateAdded = entryNode.getElementsByTagName('published').item(0).childNodes.item(0).nodeValue
        self.dateUpdated = entryNode.getElementsByTagName('updated').item(0).childNodes.item(0).nodeValue
        self.links = {}
        for n in entryNode.childNodes:
            if n.nodeName == 'link':
                linkrel = n.getAttribute('rel')
                linktype = n.getAttribute('type')
                linkhref = n.getAttribute('href')
                self.links[linkrel] = {'type': linktype, 'href': linkhref, 'rel': linkrel}
                if linkrel == 'enclosure':
                    self.links[linkrel]['title'] = n.getAttribute('title')
                    self.links[linkrel]['size'] = n.getAttribute('size')

        contentNode = entryNode.getElementsByTagName("content").item(0)
        if(contentNode):
            self.contentType = contentNode.getAttribute('type')
            self.contentZType = contentNode.getAttribute('zapi:type')

    def associateWithLibrary(self, library):
        self.libraryType = library.libraryType
        self.libraryID = library.libraryID
        self.owningLibrary = library



import logging


class Response(object):
    def __init__(self, res=None):
        logging.info("Initializing zotero.Response object")
        #info = res.info()
        self.url = res.geturl()
        #self.headers = info.dict
        self.headers = res.headers
        self.status_code = res.getcode()
        self.text = res.read()



from distutils.core import setup
setup(name='libZotero',
      version='0.2.5',
      description='Python module for interacting with the zotero.org API',
      author='Faolan Cheslack-Postava',
      url='https://github.com/fcheslack/libZotero',
      packages=['libZotero'],
      )



from exceptions import *
from library import *
from feed import *
from item import *
from items import *
from collection import *
from zcollections import *
from tag import *
from cache import *
from response import *

ZOTERO_URI = 'https://api.zotero.org'



import sys
sys.path.append('../')
import unittest
from libZotero import zotero


class TestItemCreateFunctions(unittest.TestCase):

    def setUp(self):
        self.seq = range(10)

    def test_templates(self):
        #zlib = zotero.Library('user', '475425', 'z_public_library', '')

        #create a new item of type book
        newItem = zotero.getTemplateItem('book')
        newItem.set('title', 'This is a book ---')
        newItem.set('abstractNote', 'Created using a zotero python library and the write api')

        self.assertEqual(newItem.get('title'), 'This is a book ---')
        self.assertEqual(newItem.get('abstractNote'), "Created using a zotero python library and the write api")


if __name__ == '__main__':
    unittest.main()
